"""Stimuli from Bindmann & Chubb (2004)
https://doi.org/10.1016/S0042-6989(03)00430-9

This module reproduces most of the stimuli used by Bindmann & Chubb (2004) as
they were described in that paper.

Each stimulus is provided by a separate function,
a full list can be found as stimupy.papers.bindmann2004.__all__

The output of each of these functions is a stimulus dictionary.

For a visual representation of all the stimuli and their mask,
simply run this module as a script:

    $ python stimuli/papers/bindmann2004.py

Attributes
----------
__all__ (list of str): list of all stimulus-functions
    that are exported by this module when executing
        >>> from stimupy.papers.bindmann2004 import *

References
-----------
Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
"""

import copy
import warnings

import numpy as np

from stimupy import bullseyes
from stimupy.utils.contrast_conversions import adapt_intensity_range_dict
from stimupy.utils.pad import pad_dict_by_visual_size

__all__ = [
    "bullseye_thin_gw45_gb31",
    "bullseye_thin_gw45_gb38",
    "bullseye_thin_gw45_gb45",
    "bullseye_thin_gw45_gb52",
    "bullseye_thin_gw45_gb59",
    "bullseye_thin_gw60_gb46",
    "bullseye_thin_gw60_gb53",
    "bullseye_thin_gw60_gb60",
    "bullseye_thin_gw60_gb67",
    "bullseye_thin_gw60_gb74",
    "bullseye_thin_gw75_gb61",
    "bullseye_thin_gw75_gb68",
    "bullseye_thin_gw75_gb75",
    "bullseye_thin_gw75_gb82",
    "bullseye_thin_gw75_gb89",
    "bullseye_thick_gw45_gb31",
    "bullseye_thick_gw45_gb38",
    "bullseye_thick_gw45_gb45",
    "bullseye_thick_gw45_gb52",
    "bullseye_thick_gw45_gb59",
    "bullseye_thick_gw60_gb46",
    "bullseye_thick_gw60_gb53",
    "bullseye_thick_gw60_gb60",
    "bullseye_thick_gw60_gb67",
    "bullseye_thick_gw60_gb74",
    "bullseye_thick_gw75_gb61",
    "bullseye_thick_gw75_gb68",
    "bullseye_thick_gw75_gb75",
    "bullseye_thick_gw75_gb82",
    "bullseye_thick_gw75_gb89",
]

# The authors say that they used a 15" Sony Trinitron monitor. We found the
# following specs online
monitor_height_px = 1024
monitor_width_px = 768
monitor_height_cm = 40.0
monitor_width_cm = 30.0
distance_cm = 132  # provided in the paper
ppcm = monitor_height_cm / monitor_height_px
PPD = int(np.tan(np.pi / 180.0 / 2.0) * 2.0 * distance_cm / ppcm)

TARGET_SIZE = 0.608
INTENSITY_BLACK = 1
INTENSITY_WHITE = 118
INTENSITY_BACKGROUND = 60
PAD1 = ((0, 1.2), (0, 2.7))
PAD2 = ((1.2, 0), (2.7, 0))
ORIGIN = "mean"


def gen_all(ppd=PPD, skip=False):
    stims = {}  # save the stimulus-dicts in a larger dict, with name as key
    for stim_name in __all__:
        print(f"Generating bindmann2004.{stim_name}")

        # Get a reference to the actual function
        func = globals()[stim_name]
        try:
            stim = func(ppd=ppd)

            # Accumulate
            stims[stim_name] = stim
        except NotImplementedError as e:
            if not skip:
                raise e
            # Skip stimuli that aren't implemented
            print("-- not implemented")
            pass

    return stims


def resolve_target_size(target_size=TARGET_SIZE, ppd=PPD):
    target_size_old = copy.deepcopy(target_size)
    target_size = np.round(target_size / 2 * ppd) / ppd * 2

    if target_size_old != target_size:
        warnings.warn(
            f"Rounding target_size because of ppd; {target_size_old} -> {target_size}. "
            "This might also affect stimulus size."
        )
    return target_size


def resolve_band_width(band_width, ppd=PPD):
    band_width_old = copy.deepcopy(band_width)
    band_width = np.round(band_width * ppd) / ppd

    if band_width_old != band_width:
        warnings.warn(
            f"Rounding band_width because of ppd; {band_width_old} -> {band_width}. "
            "This might also affect stimulus size."
        )
    return band_width


def bullseye_thin_gw45_gb31(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 45 & 31 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
        Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=45.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=31.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw45_gb38(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 45 & 38 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=45.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=38.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw45_gb45(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 45 & 45 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=45.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=45.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw45_gb52(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 45 & 52 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=45.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=52.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw45_gb59(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 45 & 59 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=45.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=59.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw60_gb46(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 60 & 46 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=60.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=46.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw60_gb53(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 60 & 53 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=60.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=53.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw60_gb60(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 60 & 60 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=60.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=60.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw60_gb67(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 60 & 67 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=60.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=67.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw60_gb74(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 60 & 74 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=60.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=74.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw75_gb61(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 75 & 61 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=75.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=61.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw75_gb68(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 75 & 68 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=75.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=68.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw75_gb75(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 75 & 75 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=75.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=75.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw75_gb82(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 75 & 82 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=75.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=82.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thin_gw75_gb89(ppd=PPD):
    """Bullseye thin display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.122 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 75 & 89 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.122, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=75.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=89.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw45_gb31(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 45 & 31 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=45.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=31.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw45_gb38(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 45 & 38 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=45.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=38.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw45_gb45(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 45 & 45 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=45.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=45.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw45_gb52(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 45 & 52 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=45.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=52.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw45_gb59(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 45 & 59 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=45.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=59.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw60_gb46(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 60 & 46 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=60.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=46.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw60_gb53(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 60 & 53 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=60.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=53.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw60_gb60(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 60 & 60 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=60.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=60.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw60_gb67(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 60 & 67 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=60.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=67.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw60_gb74(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 60 & 74 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=60.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=74.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw75_gb61(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 75 & 61 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=75.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=61.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw75_gb68(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 75 & 68 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=75.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=68.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw75_gb75(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 75 & 75 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=75.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=75.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw75_gb82(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 75 & 82 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=75.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=82.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


def bullseye_thick_gw75_gb89(ppd=PPD):
    """Bullseye thick display from Bindmann & Chubb (2004)
    Target size: 0.608 x 0.608 deg
    Band widths: 0.243 deg
    Target distance (center to center): 1.2 x 2.7 deg
    Target luminances: 75 & 89 cd2/m

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    -----------
    Bindmann, D. & Chubb, C. (2004). Brightness assimilation in bullseye displays.
    Vision Research, 44 (3), 309-319. https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    target_size_half = resolve_target_size(ppd=ppd) / 2
    band_width = resolve_band_width(0.243, ppd=ppd)
    radii = np.linspace(target_size_half, target_size_half + band_width * 4, 5)

    params = {
        "shape": (target_size_half + band_width * 4) * 2 * ppd - 1,
        "ppd": ppd,
        "radii": radii,
        "origin": ORIGIN,
    }

    stim1 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_BLACK, INTENSITY_WHITE),
        intensity_target=75.0,
    )
    stim2 = bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(INTENSITY_WHITE, INTENSITY_BLACK),
        intensity_target=89.0,
    )

    # Pad
    stim1 = pad_dict_by_visual_size(stim1, PAD1, ppd, pad_value=INTENSITY_BACKGROUND)
    stim2 = pad_dict_by_visual_size(stim2, PAD2, ppd, pad_value=INTENSITY_BACKGROUND)

    # Update dict keys
    stim1["img"] = np.where(stim2["pad_mask"] == 0, stim2["img"], stim1["img"])
    stim1["target_mask"] = np.where(stim2["target_mask"], 2, stim1["target_mask"])
    stim1["frame_mask2"] = stim2["frame_mask"]
    stim1["pad_mask2"] = stim2["pad_mask"]
    stim1["original_range"] = (INTENSITY_BLACK, INTENSITY_WHITE)

    # Adapt range between 0 and 1
    stim = adapt_intensity_range_dict(stim1)
    return stim


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = gen_all(skip=True)
    plot_stimuli(stims, mask=False)



import pytest

from stimupy.components import waves


@pytest.mark.parametrize(
    "ppd, length, visual_angle, n_phases, phase_width, frequency",
    (
        (32, 1024, 32, 8, 2, 1 / 4),
        (32, 1024, 32, 4, 1, 1 / 2),
        (32, 1024, 32, 8, None, 1 / 2),
        (32, 1024, 32, None, 2, None),
        (32, 1024, 32, None, None, 1 / 2),
        (None, 1024, 32, 8, 2, None),
        (None, 1024, 32, 8, 2, None),
        (None, 1024, 32, None, None, 1),
        (None, 1024, None, 8, None, 1),
    ),
)
def test_valid_params(ppd, length, visual_angle, n_phases, phase_width, frequency):
    waves.resolve_grating_params(
        ppd=ppd,
        length=length,
        visual_angle=visual_angle,
        n_phases=n_phases,
        phase_width=phase_width,
        frequency=frequency,
    )


def test_rounding():
    ppd = 36
    visual_size = (1.0, 1.0)
    frequency = 2.80
    stim = waves.square(
        ppd=ppd,
        visual_size=visual_size,
        frequency=frequency,
        period="either",
        distance_metric="horizontal",
        phase_shift=0,
        origin="corner",
        round_phase_width=True,
    )


def test_overview():
    waves.overview()



import numpy as np

__all__ = [
    "luminance2munsell",
    "munsell2luminance",
]


def luminance2munsell(lum_values, reference_white):
    """
    Transform luminance values into Munsell values.
    The luminance values do not have to correspond to specific units, as long
    as they are in the same unit as the reference white, because Munsell values
    are a perceptually uniform scale of relative luminances.

    Parameters
    ----------
    lum_values : numpy-array
    reference_white : number

    Returns
    -------
    munsell_values : numpy-array

    Reference: H. Pauli, "Proposed extension of the CIE recommendation
    on 'Uniform color spaces, color difference equations, and metric color
    terms'," J. Opt. Soc. Am. 66, 866-867 (1976)
    """

    x = lum_values / float(reference_white)
    idx = x <= (6.0 / 29) ** 3
    y1 = 841.0 / 108 * x[idx] + 4.0 / 29
    y2 = x[~idx] ** (1.0 / 3)
    y = np.empty(x.shape)
    y[idx] = y1
    y[~idx] = y2
    return 11.6 * y - 1.6


def munsell2luminance(munsell_values, reference_white):
    """
    Transform Munsell values to luminance values.
    The luminance values will be in the same unit as the reference white, which
    can be arbitrary as long as the scale is linear.

    Parameters
    ----------
    munsell_values : numpy-array
    reference_white : number

    Returns
    -------
    lum_values : numpy-array

    Reference: H. Pauli, "Proposed extension of the CIE recommendation
    on 'Uniform color spaces, color difference equations, and metric color
    terms'," J. Opt. Soc. Am. 66, 866-867 (1976)
    """
    lum_values = (munsell_values + 1.6) / 11.6
    idx = lum_values <= 6.0 / 29
    lum_values[idx] = (lum_values[idx] - 4.0 / 29) / 841 * 108
    lum_values[~idx] **= 3
    return lum_values * reference_white



import numpy as np
from PIL import Image, ImageDraw, ImageFont

from stimupy.utils import pad_dict_to_visual_size, resolution

__all__ = [
    "text",
]


def text(
    text,
    visual_size=None,
    ppd=None,
    shape=None,
    intensity_text=0.0,
    intensity_background=0.5,
    fontsize=36,
    align="center",
    # direction="ltr",
):
    """Draw given text into a (numpy) image-array

    If no shape is provided / can be resolved,
    tightly fits the bounding box of the drawn text.

    Parameters
    ----------
    text : str
        Text to draw
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees visual angle
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    intensity_text : float, optional
        intensity of text in range (0.0; 1.0), by default 0.0
    intensity_background : float, optional
        intensity value of background in range (0.0; 1.0), by default 0.5
    fontsize : int, optional
        font size, by default 36
    align : "left", "center" (default), "right"
        alignment of text, by default "center"

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the text (key: "text_mask"),
        and additional keys containing stimulus parameters
    """

    # Try to resolve resolution
    try:
        shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    except resolution.TooManyUnknownsError:
        shape = resolution.validate_shape(shape)
        visual_size = resolution.validate_visual_size(visual_size)
        ppd = resolution.validate_ppd(ppd)

    # Get font
    font = ImageDraw.ImageDraw.font
    if not font:
        try:
            # Not all machines will have Arial installed...
            font = ImageFont.truetype(
                "arial.ttf",
                fontsize,
                encoding="unic",
            )
        except OSError:
            font = ImageFont.load_default()

    # Determine dimensions of total text
    n_lines = len(text.split("\n"))
    max_length = 0
    for line in text.split("\n"):
        max_length = max(int(font.getlength(line)), max_length)
    _, top, _, bottom = font.getbbox(text)
    text_width = max_length
    text_height = int(top + bottom) * n_lines
    text_shape = (text_height, text_width)

    # Instantiate grayscale image of correct shape (in pixels)
    img = Image.new("L", (text_width, text_height), 0)
    draw = ImageDraw.Draw(img)

    # Draw text into this image
    draw.text(
        (0, 0),
        text,
        fill=1,
        font=font,
        align=align,
        # direction=direction
    )

    # Turn into mask-array
    mask = np.array(img)
    img = np.where(mask, intensity_text, intensity_background)

    # Package as dict
    stim = {
        "img": img,
        "text_mask": mask,
        "text_shape": text_shape,
    }

    # Resolve resolution
    if not shape or shape == (None, None):
        shape = text_shape
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)

    # Pad
    stim = pad_dict_to_visual_size(stim, visual_size=visual_size, ppd=ppd, pad_value=0.5)

    # Output
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": (10, 10),
        "ppd": 32,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "text(), single line": text(text="hello world", **default_params),
        "text(), multiline": text(text="hello\nworld", **default_params)
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



"""Stimuli from White (1981)
https://doi.org/10.1068/p100215

This module reproduces most of the stimuli used by White (1981) as they were
described in that paper.

Each stimulus is provided by a separate function,
a full list can be found as stimupy.papers.white1981.__all__

The output of each of these functions is a stimulus dictionary.

For a visual representation of all the stimuli and their mask,
simply run this module as a script:

    $ python stimuli/papers/white1981.py

Attributes
----------
__all__ (list of str): list of all stimulus-functions
    that are exported by this module when executing
        >>> from stimupy.papers.white1981 import *

References
-----------
White, M. (1981).
    The effect of the nature of the surround on the perceived lightness
    of grey bars within square-wave test grating.
    Perception, 10, 215-230.
    https://doi.org/10.1068/p100215
"""

import copy
import warnings

import numpy as np

import stimupy

__all__ = [
    "square_white",
    "square_black",
    "grating_white_white",
    "grating_white_black",
    "grating_black_white",
    "grating_black_black",
    "grating_white_in",
    "grating_black_in",
    "grating_white_out",
    "grating_black_out",
    "grating_white_orthogonal",
    "grating_black_orthogonal",
]

VISUAL_SIZE = 4.0  # originally 4.3592 deg
TARGET_SIZE = 0.72  # originally 0.7696 deg
BAR_WIDTH = 0.08  # originally 0.0855 deg

PPD = 50
ROTATION = 90
v1, v2, v3 = 0.0, 0.5, 1.0


def gen_all(ppd=PPD, skip=False):
    stims = {}  # save the stimulus-dicts in a larger dict, with name as key
    for stim_name in __all__:
        print(f"Generating white1981.{stim_name}")

        # Get a reference to the actual function
        func = globals()[stim_name]
        try:
            stim = func(ppd=ppd)

            # Accumulate
            stims[stim_name] = stim
        except NotImplementedError as e:
            if not skip:
                raise e
            # Skip stimuli that aren't implemented
            print("-- not implemented")
            pass

    return stims


def resolve_bar_width(bar_width=BAR_WIDTH, ppd=PPD):
    bar_width_old = copy.deepcopy(bar_width)
    bar_width = np.round(bar_width * ppd) / ppd

    if bar_width_old != bar_width:
        warnings.warn(
            f"Rounding bar_width because of ppd; {bar_width_old} -> {bar_width}. "
            "This will also effect the stimulus size."
        )
    return bar_width


def square_white(ppd=PPD):
    """A square on a white background from White (1981), Fig. 2
    Stimulus size: 4.0 x 4.0 deg (originally 4.3592 deg)
    Target size: 0.72 x 0.72 deg (originally 0.7696 deg)

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "target_size": TARGET_SIZE,
        "intensity_background": v3,
        "intensity_target": v2,
    }

    stim = stimupy.sbcs.basic(**params)
    return stim


def square_black(ppd=PPD):
    """A square on a black background from White (1981), Fig. 2
    Stimulus size: 4.0 x 4.0 deg (originally 4.3592 deg)
    Target size: 0.72 x 0.72 deg (originally 0.7696 deg)

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "target_size": TARGET_SIZE,
        "intensity_background": v1,
        "intensity_target": v2,
    }

    stim = stimupy.sbcs.basic(**params)
    return stim


def grating_white_white(ppd=PPD):
    """A white-gray grating on white background from White (1981), Fig. 2
    Stimulus size: 4.0 x 4.0 deg (originally 4.3592 deg)
    Target size: 0.72 x 0.72 deg (originally 0.7696 deg)
    Bar width: 0.08 deg (originally 0.0855 deg)

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """
    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "n_bars": 9,
        "bar_width": bar_width,
        "rotation": ROTATION,
        "intensity_background": v3,
        "intensity_bars": (v2, v3),
        "intensity_target": v2,
        "target_indices": (2, 4, 6, 8),
    }
    stim = stimupy.gratings.on_uniform(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def grating_white_black(ppd=PPD):
    """A white-gray grating on black background from White (1981), Fig. 2
    Stimulus size: 4.0 x 4.0 deg (originally 4.3592 deg)
    Target size: 0.72 x 0.72 deg (originally 0.7696 deg)
    Bar width: 0.08 deg (originally 0.0855 deg)

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """
    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "grating_size": bar_width * 9,
        "bar_width": bar_width,
        "rotation": ROTATION,
        "intensity_background": v1,
        "intensity_bars": (v2, v3),
        "intensity_target": v2,
        "target_indices": (2, 4, 6, 8),
    }
    stim = stimupy.gratings.on_uniform(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def grating_black_white(ppd=PPD):
    """A black-gray grating on white background from White (1981), Fig. 2
    Stimulus size: 4.0 x 4.0 deg (originally 4.3592 deg)
    Target size: 0.72 x 0.72 deg (originally 0.7696 deg)
    Bar width: 0.08 deg (originally 0.0855 deg)

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """
    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "grating_size": bar_width * 9,
        "bar_width": bar_width,
        "rotation": ROTATION,
        "intensity_background": v3,
        "intensity_bars": (v2, v1),
        "intensity_target": v2,
        "target_indices": (2, 4, 6, 8),
    }
    stim = stimupy.gratings.on_uniform(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def grating_black_black(ppd=PPD):
    """A black-gray grating on black background from White (1981), Fig. 2
    Stimulus size: 4.0 x 4.0 deg (originally 4.3592 deg)
    Target size: 0.72 x 0.72 deg (originally 0.7696 deg)
    Bar width: 0.08 deg (originally 0.0855 deg)

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """
    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    params = {
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "grating_size": bar_width * 9,
        "bar_width": bar_width,
        "rotation": ROTATION,
        "intensity_background": v1,
        "intensity_bars": (v2, v1),
        "intensity_target": v2,
        "target_indices": (2, 4, 6, 8),
    }
    stim = stimupy.gratings.on_uniform(**params)
    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def grating_white_in(ppd=PPD):
    """A white-gray grating on an in-phase grating from White (1981), Fig. 3
    Stimulus size: 4.08 x 4.08 deg (originally 4.3592 deg)
    Target size: 0.72 x 0.72 deg (originally 0.7696 deg)
    Bar width: 0.08 deg (originally 0.0855 deg)

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """
    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    small_params = {
        "bar_width": bar_width,
        "n_bars": 9,
        "ppd": ppd,
        "rotation": ROTATION,
        "intensity_bars": (v1, v3),
        "intensity_target": v2,
        "target_indices": (2, 4, 6, 8),
    }
    large_params = {
        "bar_width": bar_width,
        "visual_size": 51 * bar_width,
        "ppd": ppd,
        "rotation": ROTATION,
        "intensity_bars": (v3, v1),
    }
    stim = stimupy.gratings.on_grating(
        small_grating_params=small_params,
        large_grating_params=large_params,
    )

    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def grating_black_in(ppd=PPD):
    """A black-gray grating on an in-phase grating from White (1981), Fig. 3
    Stimulus size: 4.08 x 4.08 deg (originally 4.3592 deg)
    Target size: 0.72 x 0.72 deg (originally 0.7696 deg)
    Bar width: 0.08 deg (originally 0.0855 deg)

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """
    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    small_params = {
        "bar_width": bar_width,
        "n_bars": 9,
        "ppd": ppd,
        "rotation": ROTATION,
        "intensity_bars": (v3, v1),
        "intensity_target": v2,
        "target_indices": (2, 4, 6, 8),
    }
    large_params = {
        "bar_width": bar_width,
        "visual_size": 51 * bar_width,
        "ppd": ppd,
        "rotation": ROTATION,
        "intensity_bars": (v1, v3),
    }
    stim = stimupy.gratings.on_grating(
        small_grating_params=small_params,
        large_grating_params=large_params,
    )

    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def grating_white_out(ppd=PPD):
    """A white-gray grating on an out-of-phase grating from White (1981), Fig. 3
    Stimulus size: 4.0 x 4.0 deg (originally 4.3592 deg)
    Target size: 0.72 x 0.72 deg (originally 0.7696 deg)
    Bar width: 0.08 deg (originally 0.0855 deg)

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """
    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    small_params = {
        "bar_width": bar_width,
        "visual_size": (50 * bar_width, 9 * bar_width),
        "ppd": ppd,
        "rotation": -ROTATION,
        "intensity_bars": (v1, v3),
        "intensity_target": v2,
        "target_indices": (24, 26, 28, 30),
    }
    large_params = {
        "bar_width": bar_width,
        "visual_size": 50 * bar_width,
        "ppd": ppd,
        "rotation": -ROTATION,
        "intensity_bars": (v3, v1),
    }
    stim = stimupy.gratings.on_grating(
        small_grating_params=small_params,
        large_grating_params=large_params,
    )

    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def grating_black_out(ppd=PPD):
    """A black-gray grating on an out-of-phase grating from White (1981), Fig. 3
    Stimulus size: 4.0 x 4.0 deg (originally 4.3592 deg)
    Target size: 0.72 x 0.72 deg (originally 0.7696 deg)
    Bar width: 0.08 deg (originally 0.0855 deg)

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """
    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    small_params = {
        "bar_width": bar_width,
        "visual_size": (50 * bar_width, 9 * bar_width),
        "ppd": ppd,
        "rotation": -ROTATION,
        "intensity_bars": (v3, v1),
        "intensity_target": v2,
        "target_indices": (24, 26, 28, 30),
    }
    large_params = {
        "bar_width": bar_width,
        "visual_size": 50 * bar_width,
        "ppd": ppd,
        "rotation": -ROTATION,
        "intensity_bars": (v1, v3),
    }
    stim = stimupy.gratings.on_grating(
        small_grating_params=small_params,
        large_grating_params=large_params,
    )

    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def grating_white_orthogonal(ppd=PPD):
    """A white-gray grating on an orthogonal grating from White (1981), Fig. 3
    Stimulus size: 4.0 x 4.0 deg (originally 4.3592 deg)
    Target size: 0.72 x 1.36 deg (originally 0.6841 x 1.1117 deg)
    Bar width: 0.08 deg (originally 0.0855 deg)

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """
    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    small_params = {
        "visual_size": (17 * bar_width, 9 * bar_width),
        "bar_width": bar_width,
        "ppd": ppd,
        "rotation": ROTATION + 90,
        "intensity_bars": (v2, v3),
        "intensity_target": v2,
        "target_indices": (2, 4, 6, 8),
    }
    large_params = {
        "bar_width": bar_width,
        "visual_size": 51 * bar_width,
        "ppd": ppd,
        "rotation": ROTATION,
        "intensity_bars": (v3, v1),
    }
    stim = stimupy.gratings.on_grating_masked(
        small_grating_params=small_params,
        large_grating_params=large_params,
        mask_size=(9 * bar_width, 9 * bar_width, 8 * bar_width),
        mask_rotation=ROTATION,
    )

    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


def grating_black_orthogonal(ppd=PPD):
    """A black-gray grating on an orthogonal grating from White (1981), Fig. 3
    Stimulus size: 4.0 x 4.0 deg (originally 4.3592 deg)
    Target size: 0.72 x 1.36 deg (originally 0.6841 x 1.1117 deg)
    Bar width: 0.08 deg (originally 0.0855 deg)

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree.

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    White, M. (1981).
        The effect of the nature of the surround on the perceived lightness
        of grey bars within square-wave test grating.
        Perception, 10, 215-230.
        https://doi.org/10.1068/p100215
    """
    bar_width = resolve_bar_width(BAR_WIDTH, ppd)

    small_params = {
        "visual_size": (17 * bar_width, 9 * bar_width),
        "bar_width": bar_width,
        "ppd": ppd,
        "rotation": ROTATION + 90,
        "intensity_bars": (v3, v1),
        "intensity_target": v2,
        "target_indices": (2, 4, 6, 8),
    }
    large_params = {
        "bar_width": bar_width,
        "visual_size": 51 * bar_width,
        "ppd": ppd,
        "rotation": ROTATION,
        "intensity_bars": (v1, v3),
    }
    stim = stimupy.gratings.on_grating_masked(
        small_grating_params=small_params,
        large_grating_params=large_params,
        mask_size=(9 * bar_width, 9 * bar_width, 8 * bar_width),
        mask_rotation=ROTATION,
    )

    stim["target_mask"] = np.where(stim["target_mask"] != 0, 1, 0)
    return stim


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = gen_all(skip=True)
    plot_stimuli(stims, mask=False)



import numpy as np

from stimupy.components import draw_regions, mask_regions
from stimupy.components.radials import ring

__all__ = [
    "wedge",
    "segments",
]


def mask_angle(
    angles,
    rotation=0.0,
    visual_size=None,
    ppd=None,
    shape=None,
    origin="mean",
):
    """Mask a contiguous set of angles in image

    Parameters
    ----------
    angles : Sequence[float, float]
        lower- and upper-limit of angles to mask, in degrees
    rotation : float, optional
        rotation (in degrees) from 3 o'clock, counterclockwise, by default 0.0
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with boolean mask (key: "segment_mask") for pixels falling in given angle,
        and additional params
    """
    stim = mask_regions(
        edges=np.deg2rad(angles),
        distance_metric="angular",
        rotation=rotation,
        shape=shape,
        visual_size=visual_size,
        ppd=ppd,
        origin=origin,
    )
    stim["segment_mask"] = stim.pop("mask")
    return stim


def wedge(
    visual_size=None,
    ppd=None,
    shape=None,
    angle=None,
    radius=None,
    rotation=0.0,
    inner_radius=0.0,
    intensity_wedge=1.0,
    intensity_background=0.0,
    origin="mean",
):
    """Draw a wedge, i.e., segment of a disc

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    angle : float
        angular-width (in degrees) of segment
    radius : float
        radius of disc, in degrees visual angle
    rotation : float, optional
        rotation (in degrees) from 3 o'clock, counterclockwise, by default 0.0
    inner_radius : float, optional
        inner radius (in degrees visual angle), to turn disc into a ring, by default 0
    intensity_wedge : float, optional
        intensity value of wedge, by default 1.0
    intensity_background : float, optional
        intensity value of background, by default 0.0
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each segment (key: "wedge_mask"),
        and additional keys containing stimulus parameters
    """
    if angle is None:
        raise ValueError("wedge() missing argument 'angle' which is not 'None'")
    if radius is None:
        raise ValueError("wedge() missing argument 'radius' which is not 'None'")

    # Convert to inner-, outer-angle
    angles = [0, angle]

    # Draw disc
    stim = ring(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        radii=[inner_radius, radius],
        intensity_ring=intensity_wedge,
        intensity_background=intensity_background,
        origin=origin,
    )

    # Get mask for angles
    mask = mask_angle(
        rotation=rotation,
        angles=angles,
        visual_size=stim["visual_size"],
        ppd=stim["ppd"],
        shape=stim["shape"],
        origin=origin,
    )

    # Combine masks
    stim["wedge_mask"] = stim.pop("ring_mask") * mask["segment_mask"]
    stim["wedge_mask"] = np.where(stim["wedge_mask"] != 0, 1, 0).astype(int)

    # Draw image
    stim["img"] = np.where(stim["wedge_mask"], intensity_wedge, intensity_background)

    # Repackage
    stim["angle"] = angle
    stim["radius"] = radius
    stim["inner_radius"] = inner_radius
    stim["intensity_wedge"] = intensity_wedge
    stim["intensity_background"] = intensity_background
    return stim


def mask_segments(
    edges,
    rotation=0.0,
    visual_size=None,
    ppd=None,
    shape=None,
    origin="mean",
):
    """Generate mask with integer indices for consecutive angular segments

    Parameters
    ----------
    edges : Sequence[Number]
        upper-limit of each consecutive segment, in angular degrees 0-360
    rotation : float, optional
        rotation (in degrees) from 3 o'clock, counterclockwise, by default 0.0
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each segment (key: "segment_mask"),
        and additional keys containing stimulus parameters
    """
    stim = mask_regions(
        distance_metric="angular",
        edges=np.deg2rad(edges),
        rotation=rotation,
        shape=shape,
        visual_size=visual_size,
        ppd=ppd,
        origin=origin,
    )
    stim["segment_mask"] = stim.pop("mask")
    return stim


def segments(
    visual_size=None,
    ppd=None,
    shape=None,
    angles=None,
    rotation=0.0,
    intensity_background=0.5,
    intensity_segments=(0.0, 1.0),
    origin="mean",
):
    """Generate mask with integer indices for sequential angular segments

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    angles : Sequence[Number] or None (default)
        upper-limit of each segment, in angular degrees 0-360
    rotation : float, optional
        rotation (in degrees) from 3 o'clock, counterclockwise, by default 0.0
    intensity_background : Number
        intensity value for background; default is 0.5.
    intensity_segments : Sequence[Number, ...]
        intensity value for each segment, from inside to out.
        If fewer intensities are passed than number of radii, cycles through intensity_segments
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)angles
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each segment (key: "segment_mask"),
        and additional keys containing stimulus parameters
    """
    if angles is None:
        raise ValueError("segments() missing argument 'angles' which is not 'None'")
    if not isinstance(angles, (int, float)):
        if np.diff(angles).min() < 0:
            raise ValueError("angles need to monotonically increase")

    # Get mask
    stim = mask_segments(
        edges=angles,
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rotation=rotation,
        origin=origin,
    )

    # Draw image
    stim["img"] = draw_regions(
        stim["segment_mask"],
        intensities=intensity_segments,
        intensity_background=intensity_background,
    )

    # Update args
    stim["angles"] = angles
    stim["intensity_background"] = intensity_background
    stim["intensity_segments"] = intensity_segments

    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": (10, 10),
        "ppd": 20,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "angulars_wedge": wedge(**default_params, angle=30, radius=4),
        "angulars_segments": segments(**default_params, angles=(30, 120, 300)),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import numpy as np

from stimupy.components.shapes import disc, rectangle
from stimupy.stimuli import bullseyes
from stimupy.utils import make_two_sided, pad_by_visual_size, pad_to_shape, resolution

__all__ = [
    "generalized",
    "basic",
    "square",
    "circular",
    "with_dots",
    "dotted",
    "basic_two_sided",
    "square_two_sided",
    "circular_two_sided",
    "with_dots_two_sided",
    "dotted_two_sided",
]


def generalized(
    visual_size=None,
    ppd=None,
    shape=None,
    target_size=None,
    target_position=None,
    intensity_background=0.0,
    intensity_target=0.5,
    rotation=0.0,
):
    """Simultaneous contrast stimulus with free target placement

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    target_size : float or (float, float)
        size [height, width] of the target, in degrees visual angle
    target_position : float or (float, float)
        position of the target in degree visual angle (height, width);
        if None, place target centrally
    intensity_background : float, optional
        intensity value for background, by default 0.0
    intensity_target : float, optional
        intensity value for target, by default 0.5
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)


    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Chevreul, M. (1855).
        The principle of harmony and contrast of colors.
    """
    if target_size is None:
        raise ValueError("generalized() missing argument 'target_size' which is not 'None'")

    stim = rectangle(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rectangle_size=target_size,
        rectangle_position=target_position,
        intensity_background=intensity_background,
        intensity_rectangle=intensity_target,
        rotation=rotation,
    )

    stim["target_mask"] = stim.pop("rectangle_mask")
    stim["target_size"] = stim.pop("rectangle_size")
    stim["target_position"] = stim.pop("rectangle_position")
    stim["intensity_target"] = stim.pop("intensity_rectangle")

    return stim


def basic(
    visual_size=None,
    ppd=None,
    shape=None,
    target_size=None,
    intensity_background=0.0,
    intensity_target=0.5,
):
    """Simultaneous contrast stimulus with central target

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    target_size : float or (float, float)
        size [height, width] of the target, in degrees visual angle
    intensity_background : float, optional
        intensity value for background, by default 0.0
    intensity_target : float, optional
        intensity value for target, by default 0.5

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Chevreul, M. (1855).
        The principle of harmony and contrast of colors.
    """
    if target_size is None:
        raise ValueError("basic() missing argument 'target_size' which is not 'None'")

    stim = generalized(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        target_size=target_size,
        target_position=None,
        intensity_background=intensity_background,
        intensity_target=intensity_target,
    )
    return stim


def square(
    target_radius,
    visual_size=None,
    ppd=None,
    shape=None,
    surround_radius=None,
    rotation=0.0,
    intensity_surround=0.0,
    intensity_background=0.5,
    intensity_target=0.5,
    origin="mean",
):
    """Simultaneous contrast stimulus with square target and surround field

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    target_radius : float
        radius of target, in degrees visual angle
    surround_radius : float
        radius of surround context field, in degrees visual angle
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    intensity_surrond : float, optional
        intensity of surround context field, by default 0.0
    intensity_background : float, optional
        intensity value of background, by default 0.5
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each frame (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    stim = bullseyes.rectangular_generalized(
        radii=(target_radius, surround_radius),
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        rotation=rotation,
        intensity_frames=(0.0, intensity_surround),
        intensity_background=intensity_background,
        intensity_target=intensity_target,
        origin=origin,
    )
    return stim


def circular(
    target_radius,
    visual_size=None,
    ppd=None,
    shape=None,
    surround_radius=None,
    intensity_surround=0.0,
    intensity_background=0.5,
    intensity_target=0.5,
    origin="mean",
):
    """Simultaneous contrast stimulus with circular target and surround field

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    target_radius : float
        radius of target, in degrees visual angle
    surround_radius : float
        radius of surround context field, in degrees visual angle
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    intensity_surrond : float, optional
        intensity of surround context field, by default 0.0
    intensity_background : float (optional)
        intensity value of background, by default 0.5
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each frame (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    stim = bullseyes.circular_generalized(
        radii=(target_radius, surround_radius),
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        intensity_rings=(0.0, intensity_surround),
        intensity_background=intensity_background,
        intensity_target=intensity_target,
        origin=origin,
    )
    return stim


def with_dots(
    visual_size=None,
    ppd=None,
    shape=None,
    n_dots=None,
    dot_radius=None,
    distance=None,
    target_shape=None,
    intensity_background=0.0,
    intensity_dots=1.0,
    intensity_target=0.5,
):
    """Simultaneous contrast stimulus with dots

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    n_dots : int or (int, int)
        stimulus size defined as the number of dots in y and x-directions
    dot_radius : float
        radius of dots
    distance : float
        distance between dots in degree visual angle
    target_shape : int or (int, int)
        target shape defined as the number of dots that fit into the target
    intensity_background : float
        intensity value for background
    intensity_dots : float
        intensity value for dots
    intensity_target : float
        intensity value for target

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Bressan, P., & Kramer, P. (2008).
        Gating of remote effects on lightness.
        Journal of Vision, 8(2), 16-16.
        https://doi.org/10.1167/8.2.16
    """
    if n_dots is None:
        raise ValueError("with_dots() missing argument 'n_dots' which is not 'None'")
    if dot_radius is None:
        raise ValueError("with_dots() missing argument 'dot_radius' which is not 'None'")
    if distance is None:
        raise ValueError("with_dots() missing argument 'distance' which is not 'None'")
    if target_shape is None:
        raise ValueError("with_dots() missing argument 'target_shape' which is not 'None'")

    # n_dots = number of dots vertical, horizontal, analogous to degrees
    n_dots = resolution.validate_visual_size(n_dots)

    if shape is None and visual_size is None:
        visual_size = (
            n_dots[0] * dot_radius * 2 + n_dots[0] * distance,
            n_dots[1] * dot_radius * 2 + n_dots[1] * distance,
        )

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    # target shape = is in number of dots
    target_shape = resolution.validate_visual_size(target_shape)

    # Figure out pixels_per_dot
    padding = (distance / 2.0,)
    patch = disc(
        radius=dot_radius,
        ppd=ppd,
        intensity_background=0.0,
        intensity_disc=1,
    )["img"]
    patch = pad_by_visual_size(img=patch, padding=padding, ppd=ppd, pad_value=0.0)
    pixels_per_dot = patch.shape
    patch = np.tile(patch, (int(n_dots[0]), int(n_dots[1])))

    # Target shape = target n_dots * pixels_per_dot
    rect_shape = resolution.shape_from_visual_size_ppd(
        visual_size=target_shape, ppd=pixels_per_dot
    )
    rect_visual_size = resolution.visual_size_from_shape_ppd(shape=rect_shape, ppd=ppd)

    try:
        patch = pad_to_shape(patch, shape, 0)
    except Exception:
        raise ValueError("visual_size or shape_argument are too small. Advice: set to None")

    # Create the sbc in the background:
    img_shape = patch.shape

    sbc = rectangle(
        shape=img_shape,
        ppd=ppd,
        rectangle_size=rect_visual_size,
        intensity_background=intensity_background,
        intensity_rectangle=intensity_target,
    )

    img = np.where(patch == 1, intensity_dots, sbc["img"])
    mask = np.where(patch == 1, 0, sbc["rectangle_mask"])

    stim = {
        "img": img,
        "target_mask": mask.astype(int),
        "shape": shape,
        "visual_size": visual_size,
        "ppd": ppd,
        "n_dots": n_dots,
        "dot_radius": dot_radius,
        "distance": distance,
        "target_shape": target_shape,
        "intensity_background": intensity_background,
        "intensity_dots": intensity_dots,
        "intensity_target": intensity_target,
    }

    return stim


def dotted(
    visual_size=None,
    ppd=None,
    shape=None,
    n_dots=None,
    dot_radius=None,
    distance=None,
    target_shape=None,
    intensity_background=0.0,
    intensity_dots=1.0,
    intensity_target=0.5,
):
    """Dotted simultaneous contrast

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of grating, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of grating, in pixels
    n_dots : int or (int, int)
        stimulus size defined as the number of dots in y and x-directions
    dot_radius : float
        radius of dots
    distance : float
        distance between dots in degree visual angle
    target_shape : int or (int, int)
        target shape defined as the number of dots that fit into the target
    intensity_background : float
        intensity value for background
    intensity_dots : float
        intensity value for dots
    intensity_target : float
        intensity value for target

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for the targets (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Bressan, P., & Kramer, P. (2008).
        Gating of remote effects on lightness.
        Journal of Vision, 8(2), 16-16.
        https://doi.org/10.1167/8.2.16
    """
    if n_dots is None:
        raise ValueError("dotted() missing argument 'n_dots' which is not 'None'")
    if dot_radius is None:
        raise ValueError("dotted() missing argument 'dot_radius' which is not 'None'")
    if distance is None:
        raise ValueError("dotted() missing argument 'distance' which is not 'None'")
    if target_shape is None:
        raise ValueError("dotted() missing argument 'target_shape' which is not 'None'")

    # n_dots = number of dots vertical, horizontal, analogous to degrees
    n_dots = resolution.validate_visual_size(n_dots)

    if shape is None and visual_size is None:
        visual_size = (
            n_dots[0] * dot_radius * 2 + n_dots[0] * distance,
            n_dots[1] * dot_radius * 2 + n_dots[1] * distance,
        )

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x and y direction")

    # target shape = is in number of dots
    target_shape = resolution.validate_visual_size(target_shape)

    # Figure out pixels_per_dot
    padding = (distance / 2.0,)
    patch = disc(
        radius=dot_radius,
        ppd=ppd,
        intensity_background=0.0,
        intensity_disc=1.0,
    )["img"]
    patch = pad_by_visual_size(img=patch, padding=padding, ppd=ppd, pad_value=0.0)
    pixels_per_dot = patch.shape
    patch = np.tile(patch, (int(n_dots[0]), int(n_dots[1])))

    # Target shape = target n_dots * pixels_per_dot
    rect_shape = resolution.shape_from_visual_size_ppd(
        visual_size=target_shape, ppd=pixels_per_dot
    )
    rect_visual_size = resolution.visual_size_from_shape_ppd(shape=rect_shape, ppd=ppd)

    try:
        patch = pad_to_shape(patch, shape, 0)
    except Exception:
        raise ValueError("visual_size or shape_argument are too small. Advice: set to None")

    # Create the sbc and img:
    img_shape = patch.shape
    sbc = rectangle(
        shape=img_shape,
        ppd=ppd,
        rectangle_size=rect_visual_size,
        intensity_background=intensity_background,
        intensity_rectangle=intensity_target,
    )

    img = np.where(patch, intensity_dots, intensity_background)
    img = np.where(patch + sbc["rectangle_mask"] == 2, intensity_target, img)
    mask = np.where(patch + sbc["rectangle_mask"] == 2, 1, 0)

    stim = {
        "img": img,
        "target_mask": mask.astype(int),
        "shape": shape,
        "visual_size": visual_size,
        "ppd": ppd,
        "n_dots": n_dots,
        "dot_radius": dot_radius,
        "distance": distance,
        "target_shape": target_shape,
        "intensity_background": intensity_background,
        "intensity_dots": intensity_dots,
        "intensity_target": intensity_target,
    }

    return stim


generalized_two_sided = make_two_sided(
    generalized,
    two_sided_params=(
        "target_size",
        "target_position",
        "rotation",
        "intensity_target",
        "intensity_background",
    ),
)


basic_two_sided = make_two_sided(
    basic, two_sided_params=("target_size", "intensity_target", "intensity_background")
)


square_two_sided = make_two_sided(
    square,
    two_sided_params=(
        "target_radius",
        "surround_radius",
        "rotation",
        "intensity_target",
        "intensity_surround",
        "intensity_background",
    ),
)


circular_two_sided = make_two_sided(
    circular,
    two_sided_params=(
        "target_radius",
        "surround_radius",
        "intensity_target",
        "intensity_surround",
        "intensity_background",
    ),
)


with_dots_two_sided = make_two_sided(
    with_dots, two_sided_params=("intensity_dots", "intensity_background", "intensity_target")
)


dotted_two_sided = make_two_sided(
    dotted, two_sided_params=("intensity_target", "intensity_background", "intensity_dots")
)


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "ppd": 30,
    }
    default_params.update(kwargs)

    dot_params = {
        "n_dots": 5,
        "dot_radius": 2,
        "distance": 0.05,
        "target_shape": 3,
    }

    # fmt: off
    stimuli = {
        "sbc_generalized": generalized(**default_params, visual_size=10, target_size=(3, 4), target_position=(1, 2)),
        "sbc_basic": basic(**default_params, visual_size=10, target_size=3),
        "sbc_square": square(**default_params, visual_size=10, target_radius=1, surround_radius=2),
        "sbc_circular": circular(**default_params, visual_size=10, target_radius=1, surround_radius=2),
        "sbc_with_dots": with_dots(**default_params, **dot_params),
        "sbc_dotted": dotted(**default_params, **dot_params),

        "sbc_2sided": basic_two_sided(**default_params, visual_size=10, target_size=2, intensity_background=(0.0, 1.0)),
        "sbc_square_2sided": square_two_sided(**default_params, visual_size=10, target_radius=1, surround_radius=2, intensity_surround=(1.0, 0.0)),
        "sbc_circular_2sided": circular_two_sided(**default_params, visual_size=10, target_radius=1, surround_radius=2, intensity_surround=(1.0, 0.0)),
        "sbc_with_dots_2sided": with_dots_two_sided(**default_params, **dot_params, intensity_background=(0.0, 1.0), intensity_dots=(1.0, 0.0)),
        "sbc_dotted_2sided": dotted_two_sided(**default_params, **dot_params, intensity_background=(0.0, 1.0), intensity_dots=(1.0, 0.0)),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



from stimupy.components import waves
from stimupy.components.gaussians import gaussian

__all__ = ["gabor"]


def gabor(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensities=(0.0, 1.0),
    origin="center",
    round_phase_width=False,
    sigma=None,
):
    """Draw a Gabor: a sinewave grating in a Gaussian envelope

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_bars : Number, or None (default)
        number of bars in the grating
    bar_width : Number, or None (default)
        width of a single bar, in degrees visual angle
    sigma : float or (float, float)
        sigma of Gaussian in degree visual angle (y, x)
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensities : Sequence[float, ...]
        maximal intensity value for each bar, by default (0.0, 1.0).
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each bar (key: "grating_mask"),
        and additional keys containing stimulus parameters
    """
    if sigma is None:
        raise ValueError("gabor() missing argument 'sigma' which is not 'None'")

    stim = waves.sine(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_bars,
        phase_width=bar_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        intensities=intensities,
        origin=origin,
        distance_metric="oblique",
        round_phase_width=round_phase_width,
    )

    gaussian_window = gaussian(
        visual_size=visual_size,
        ppd=ppd,
        sigma=sigma,
        origin=origin,
    )
    mean_int = (intensities[0] + intensities[1]) / 2
    stim["img"] = (stim["img"] - mean_int) * gaussian_window["img"] + mean_int
    del stim["intensities"]

    return {
        **stim,
        "sigma": sigma,
        "gaussian_mask": gaussian_window["gaussian_mask"],
        "intensities": intensities,
    }


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 10,
        "ppd": 20,
        "rotation": 45,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "gabor": gabor(**default_params, frequency=1, sigma=2, phase_shift=0, round_phase_width=False, origin="center"),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)
